"""
A Python API for Bitfrog notifications.
"""

import urllib.parse
import re
import requests

ENDPOINT = "http://localhost:9000/v1"

class TokenException(Exception):
    """A project token Exception"""

class ChannelException(Exception):
    """A channel Exception"""

class ProjectException(Exception):
    """A project exception."""

def is_valid_token(token: str) -> bool:
    """
    Check if a tokens format is valid or not.

    >>> is_valid_token("a410-c45d-bce6-60af")
    True
    >>> is_valid_token("this is a token")
    False

    Args:
        token (str): The project token.
    
    Returns:
        bool: Whether the token is valid.
    """
    return re.fullmatch(r"^[\da-fA-F]{4}-[\da-fA-F]{4}-[\da-fA-F]{4}-[\da-fA-F]{4}$", token) is not None

def ping(timeout: requests.Timeout | None = None) -> bool:
    """Pings the server to check if it's responding.

    Args:
        timeout (Timeout, optional): The timeout for the request.

    Returns:
        bool: Whether the server responds or not.
    """
    r = requests.get(ENDPOINT + "/ping", timeout=timeout)

    return r.ok

def notify(message: str, token: str, title: str = None, 
           channel: str = None, timeout: requests.Timeout | None = None) -> None:
    """
    Send a quick notification given a message and project token.

    *Its recomended to use `Project.notify()` or `Channel.notify()` instead.*

    Args:
        message (str): The message to send.
        token (str): The project token.
        title (str, optional): The title of the notification.
        channel (str, optional): The name of the channel to send to.
        timeout (Timeout, optional): The timeout for the request.
    """
    if not is_valid_token(token):
        raise TokenException("The token provided is not correctly formatted.")

    safe_message = urllib.parse.quote_plus(message)
    safe_token = urllib.parse.quote_plus(token)
    safe_title = urllib.parse.quote_plus(title)
    safe_channel = urllib.parse.quote_plus(channel)

    url = f"{ENDPOINT}/notify?token={safe_token}&message={safe_message}"
    if title is not None :
        url += f"&title={safe_title}"
    if channel is not None:
        url += f"&title={safe_channel}"

    requests.get(url, timeout=timeout)

class Channel:
    """
    A class representing a project channel.

    >>> channel = Channel("xxxx-xxxx-xxxx-xxxx", "Test Channel")
    >>> channel.notify("Hello World!")
    """
    def __init__(self, project_token: str, name: str, check=False):
        """
        Args:
            project_token (str): The project token.
            name (str): The channel name.
            check (bool): Whether to check if the channel exists.
        """
        if not is_valid_token(project_token):
            raise TokenException("The token provided is not correctly formatted.")

        self.project_token = project_token
        self.name = name

        if check:
            raise ChannelException("This channel does not exist.")

    def notify(self, message: str, title: str = None, timeout: requests.Timeout | None = None):
        """
        Send a notification to a channel.

        Args:
            message (str): The message to send.
            title (str, optional): The title of the notification.
            timeout (Timeout, optional): The timeout for the request.
        """
        notify(message, self.project_token, title=title, channel=self.name, timeout=timeout)

class Project:
    """
    A class representing a project.
    
    >>> project = Project("xxxx-xxxx-xxxx-xxxx")
    >>> project.notify("Hello World!") # Notifies the primary channel.
    
    """
    def __init__(self, token, check=False):
        """
        Args:
            token (str): The project token.
            check (bool): Whether to check if the project exists.
        """
        if not is_valid_token(token):
            raise TokenException("The token provided is not correctly formatted.")

        self.token = token

        if check:
            raise ProjectException("This project does not exist.")

    def channel(self, name: str, check=False) -> Channel:
        """
        Get a channel from a project.

        >>> channel = project.channel("Test Channel")
        >>> channel.notify("Hello World!")

        Args:
            name (str): The channel name.
            check (bool): Whether to check if the channel exists.
        
        Returns:
            Channel: The channel requested.
        """
        if check:
            raise ChannelException("This channel does not exist.")

        return Channel(self.token, name)
    
    def notify(self, message: str, title: str = None, channel: str = None,
               timeout: requests.Timeout | None = None):
        """
        Send a notification to a project.

        Args:
            message (str): The message to send.
            title (str, optional): The title of the notification.
            channel (str, optional): The name of the channel to send to.
            timeout (Timeout, optional): The timeout for the request.
        """
        notify(message, self.token, title=title, channel=channel, timeout=timeout)
